package com.netapp.platform.twiddle;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.MBeanInfo;
import javax.management.MBeanOperationInfo;
import javax.management.MBeanParameterInfo;
import javax.management.MBeanServerConnection;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;

public class JmxExecutor {

	private final String urlString = "service:jmx:rmi:///jndi/rmi://localhost:33333/jmxrmi";
	private JMXConnector jmxConnector = null;
	private MBeanServerConnection connection = null;

	public JmxExecutor(String serverUrl) throws Exception {
		init(serverUrl);
	}

	private void init(String serverUrl) throws Exception {
		JMXServiceURL serviceURL = null;
		if (serverUrl == null || serverUrl.isEmpty()) {
			serviceURL = new JMXServiceURL(urlString);
		} else {
			serviceURL = new JMXServiceURL(serverUrl);
		}

		jmxConnector = JMXConnectorFactory.connect(serviceURL, null);
		connection = jmxConnector.getMBeanServerConnection();
	}

	public String invoke(String[] commandArgs) throws Exception {

	    String name = commandArgs[0];
        String operationName = commandArgs[1];

        // Fetch the list of available operations
        ObjectName mbeanName = new ObjectName(name);
        MBeanInfo mbeanInfo = connection.getMBeanInfo(mbeanName);
        MBeanOperationInfo[] opertaionsInfo = mbeanInfo.getOperations();

        // Find out the matching operation out of all the operations available
        String[] argTypes = null;
        for (int i = 0; i < opertaionsInfo.length; i++) {
            if(opertaionsInfo[i].getName().equalsIgnoreCase(operationName)) {
                MBeanParameterInfo[] paramInfo = opertaionsInfo[i].getSignature();
                argTypes = new String[paramInfo.length];
                for (int p = 0; p < paramInfo.length; p++) {
                    argTypes[p] = paramInfo[p].getType();
                }
            }
        }

        // Copy the values for the operation's arguments from command line arguments
        int position = 2;
        String[] values = new String[commandArgs.length - position];
        System.arraycopy(commandArgs, position, values, 0, commandArgs.length - position);

        if (values.length != argTypes.length) {
            System.out.println("Operation " + operationName + " takes " + argTypes.length + " arguments but you have provided " + values.length);
            return null;
        }

        // Construct array of typed arguments for the operation
        Object[] typedArgs = new Object[argTypes.length];
        for (int p = 0; p < typedArgs.length; p++) {
            typedArgs[p] = getTypedArg(argTypes[p], values[p]);
        }

        Object result = connection.invoke(mbeanName, operationName, typedArgs, argTypes);

		return result != null ? result.toString() : null;
	}

	private Object getTypedArg(String clazz, String value) {

        if(clazz.equalsIgnoreCase(Boolean.class.getName()) ||
                clazz.equalsIgnoreCase(boolean.class.getName())) {
            return Boolean.parseBoolean(value);
        }
        if(clazz.equalsIgnoreCase(Byte.class.getName()) ||
                clazz.equalsIgnoreCase(byte.class.getName())) {
	        return Byte.parseByte(value);
        }
        if(clazz.equalsIgnoreCase(Short.class.getName()) ||
                clazz.equalsIgnoreCase(short.class.getName())) {
	        return Short.parseShort(value);
        }
        if(clazz.equalsIgnoreCase(Integer.class.getName()) ||
                clazz.equalsIgnoreCase(int.class.getName())) {
	        return Integer.parseInt(value);
	    }
        if(clazz.equalsIgnoreCase(Long.class.getName()) ||
                clazz.equalsIgnoreCase(long.class.getName())) {
            return Long.parseLong(value);
        }
        if(clazz.equalsIgnoreCase(Float.class.getName()) ||
                clazz.equalsIgnoreCase(float.class.getName())) {
	        return Float.parseFloat(value);
	    }
        if(clazz.equalsIgnoreCase(Double.class.getName()) ||
                clazz.equalsIgnoreCase(double.class.getName())) {
	        return Double.parseDouble(value);
	    }

        // We should not end up here.
        return value;
	}

    public String get(String[] commandArgs) throws Exception {

		ObjectName mbeanName = new ObjectName(commandArgs[0]);

        /* commandArgs[0] is the mbeanName but getAttributes() will silently
         * fail to find an attribute of that name and effectively skip over it
         * so we can safely pass it in as a parameter here.  getAttributes()
         * will silently fail and skip over any non-existent attributes as
         * well so we need only build up a result string of found attributes,
         * if any.
         */
        AttributeList list = connection.getAttributes(mbeanName, commandArgs);
        if (list.asList().isEmpty()) {
            return null;
        }

        StringBuffer result = new StringBuffer();
        boolean firstTime = true;
        for (Attribute a : list.asList()) {
            // Don't prepend newline on the first iteration through the list
            if (!firstTime) {
                result.append("\n");
            } else {
                firstTime = false;
            }
            result.append(a.getName() + "=" + a.getValue().toString());
        }
        return result.toString();
	}

	public void destroy() throws Exception {
		jmxConnector.close();
	}
}
